/******************************************************************************\
**                                                                            **
** Licensed Materials - Property of IBM                                       **
** IBM Toolkit for MPEG-4                                                     **
** Copyright IBM Corp. 1998-2003  All Rights Reserved                         **
**                                                                            **
\******************************************************************************/

import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Rectangle;

import player.PlayerAdapter;
import player.PlayerControl;
import player.PlayerControlPanel;
import player.PlayerControlPanelFactory;
import player.PlayerFactory;

/*******************************************************************************
 * MPEG4 Applet Sample
 *
 * Applet parameters
 *  "url"           = {The source url}
 *  "urlBase"       =  "Codebase" | ["Docbase"]                 <Used when url is relative>
 *  "panel"         = ["None"] | "Mini" | "Basic" | "Default"   <Control panel type>
 *  "panelBGColor"  = "RRGGBB"                                  <Background color for panels>
 *  "showCanvas"    = ["true"] | "false"                        <Can keep canvas hidden if audio only>
 *  "autoStart"     = ["true"] | "false"                        <Content will be opened AND played if true>
 *  "scaling"       = "true" | ["false"]                        <Content will be scaled if true>
 *  "atDuration"    = ["Continue"] | "Stop" |  "Repeat"
 *
 * The url param above must be supply the content reference, such as "myFile.m4x".
 * The panelBGColor is a hex number as per HTML colors and may be preceeded by a #, thus the RRGGBB
 *   must be a value such as "FF0000" (which is red) or "808080" etc (gray). "#ffff00" is yellow.
 * The other parameters have values are as listed above with the value in [] being the default.
 * Text in <> is just comments.
 *
 */
public class M4AppletSample extends Applet implements Runnable
{
  private final static String kPlayerName = "IBM Toolkit for MPEG-4 Sample Applet";

  private PlayerControl      m4Player;     // This is the control interface for the player
  private PlayerAdapter      m4Listen;     // the listener class

  private boolean playerAdded;             // True when player renderer component is added to our applet

  private String   url;                    // A url for the content

  private boolean  useCodeBase;            // Set true if urlBase=Codebase
  private String   usePanel;               // Which panel to use; null => no panel
  private Color    bgColor;                // Background color for panel; null => do not override panel background color

  private boolean  showCanvas = true;
  private boolean  autoStart  = true;
  private boolean  scaling;
  private int      endAction = PlayerControl.END_ACTION_CONTINUE;

  private Thread   theThread;
  private boolean  stopThread;

  /*****************************************************************************
   * Constructor for our sample player applet. Set a border layout so the main
   * renderer component can be put in the center and any control panel at the
   * bottom (south).
   */
  public M4AppletSample()
  {
    setLayout(new BorderLayout());
  }

  /*****************************************************************************
   * Overrides default applet init() method. Read in the various applet
   * parameters
   */
  public void init()
  {
    url = getParameter("url");                // url is the main url for the content
    //
    String base  = getParameter("urlBase");   // Code or doc base for urls
    if (base != null && base.equalsIgnoreCase("codebase")) useCodeBase = true;

    String ctlPanel = getParameter("panel");  // Which control panel to use if any
    if (ctlPanel != null)
    {
      // Allow case independent form of panel selection. In case we have more panels then just store the
      // string but then that must match in case exactly to the PlayerControlPanelFactory's key.
      //
      if      (ctlPanel.equalsIgnoreCase("mini"))    usePanel = "Mini";
      else if (ctlPanel.equalsIgnoreCase("basic"))   usePanel = "Basic";
      else if (ctlPanel.equalsIgnoreCase("default")) usePanel = "Default";
      else if (ctlPanel.equalsIgnoreCase("none"))    usePanel = null;
      else usePanel = ctlPanel;
    }
    //
    String ctlColor    = getParameter("panelBGColor"); // Background color for panel
    if (ctlColor != null)
    {
      if (ctlColor.startsWith("#") && ctlColor.length() > 1)
      {
        ctlColor = ctlColor.substring(1);
        ctlColor.trim();
      }
      try {bgColor = new Color(Integer.parseInt(ctlColor, 16)); }
      catch (NumberFormatException e) {}
    }

    String showCanvas = getParameter("showCanvas");
    if (showCanvas != null && showCanvas.equalsIgnoreCase("false")) this.showCanvas = false;

    String autoStart   = getParameter("autoStart");
    if (autoStart != null && autoStart.equalsIgnoreCase("false")) this.autoStart = false;

    String scale       = getParameter("scaling");
    if (scale != null && scale.equalsIgnoreCase("true")) scaling = true;

    String atDuration  = getParameter("atDuration");
    if (atDuration != null)
    {
      if      (atDuration.equalsIgnoreCase("continue")) endAction = PlayerControl.END_ACTION_CONTINUE;
      else if (atDuration.equalsIgnoreCase("stop"))     endAction = PlayerControl.END_ACTION_STOP;
      else if (atDuration.equalsIgnoreCase("repeat"))   endAction = PlayerControl.END_ACTION_REPEAT;
    }

    // Create an instance of PlayerControl on which we can open the content etc
    //
    m4Player = PlayerFactory.createMPEG4Player();

    m4Listen = new M4AppletSampleAdapter();     // Set up our listener
    m4Player.addListener(m4Listen);             // and register it
    m4Player.setAutoSize(false);
    m4Player.setScaling(scaling);
    m4Player.setPlayerEndAction(endAction);

    createAbsoluteURL();                        // Make sure we have absolute url for our player

    if (usePanel != null)                       // Add a panel if param indicates to do so
    {
      PlayerControlPanel pcp = PlayerControlPanelFactory.createPlayerControlPanel(usePanel, m4Player);
      if (pcp != null)
      {
        if (bgColor != null)                    // And we can set its background color to something
        {                                       // perhaps that blends better with the content - but
          pcp.setBackground(bgColor);           // note that enough contrast is required to see
        }                                       // any controls
        add("South", pcp);
        pcp.attachToPlayer();
      }
    }
  }

  /*****************************************************************************
   * Overrides default applet start() method. We start a thread to do the
   * loading of the content as that may take some time and we don't want to
   * hold up the start.
   */
  public void start()
  {
    theThread = new Thread(this, "M4SampleApplet loader");
    theThread.start();
  }

  /*****************************************************************************
   * Overrides default applet stop() method. Here we stop the thread and once
   * the thread is ended then we stop the player. The thread may have already
   * ended - that would be the normal case - stopping the playback here is just
   * in case its still running - and the stopUrl method ensures the content
   * playback is stopped and the content is closed regardless of player state.
   */
  public void stop()
  {
    stopThread = true;

    if (theThread != null)
    {
      while (theThread.isAlive())
      {
        try {theThread.join(50);} catch (InterruptedException e) {/*nothing*/}
      }
    }

    theThread  = null;
    stopThread = false;

    m4Player.stopUrl();
  }

  /*****************************************************************************
   * Overrides default applet destroy() method. We can remove the listener from
   * the player but since its all going to be garbage collected now it really
   * doesn't have to be done.
   */
  public void destroy()
  {
    m4Player.removeListener(m4Listen);
  }

  /*****************************************************************************
   * The thread starts the player in the background.
   */
  public void run()
  {
    // A delay while we show our sample logo title. If the user goes off the
    // page before we have completed the delay in this thread then we detect
    // the stopThread boolean which is set in stop() and discontinue
    //
    for (int i=0; i < 80 && !stopThread; i++)
    {
      try {Thread.sleep(50);} catch (InterruptedException e) {/*nothing*/}
    }

    if (!stopThread)                          // Do not do this is we should stop
    {
      if (url != null && url.length() > 0)    //  If a url has been supplied
      {                                       //  the load it
        loadIt(url);
      }
    }
  }

  /*****************************************************************************
   * Loads given content via the URL into the player. This is done on a
   * separate thread kicked off from the start() method here. If applet stop()
   * is called before this is done it waits for this thread to complete anyway
   * so there is no conflict between this thread and the stop() context here in
   * shutting down the applet.
   *
   * @param url    The string for the URL for the content
   */
  private void loadIt(String url)
  {
    m4Player.stopUrl(); // Stop & close MPEG-4 player (should not be needed here really)
                        // but just in case there was a problem its safe to do this
    try
    {                   // For audio only content we don't need the renderer component
      if (showCanvas)   // but not adding component must be set via the param
      {
        add("Center", m4Player.getRendererComponent());
        playerAdded = true;
        validate();
      }

      showStatus("Opening media " + url);
      m4Player.open(url);                     // Open the content and if
      if (autoStart) m4Player.start();        // if autostarting it then start playing too
    }
    catch (Exception e)
    {
      if (showCanvas)
      {
        remove(m4Player.getRendererComponent());
        validate();
      }
      playerAdded = false;
      m4Player.stopUrl();                     // Stop MPEG-4 player & show error in status bar
      String errMsg = "Problem opening media " + e;
      showStatus(errMsg);
      System.err.println(errMsg);             // Also put it to the error console
    }
  }

  /*****************************************************************************
   * Creates an absolute URL if the reference given was absolute then it is
   * not altered, only relative URLs are made absolute being prepended with
   * codebase or docbase paths according to the useCodebase param
   */
  private void createAbsoluteURL()
  {
    // if the URL is a relative file, prepend code or doc base. If the url contains a protocol
    // then this is used as is for the url e.g. "rtsp://myserver.com/mycontent.mp4"
    //
    if (url != null)
    {
      int index = url.indexOf("://");          // Does it have a protocol spec
      if (index == -1 && !isAbsolute(url))     // If no protocol spec and relative prepend base
      {
        String base = useCodeBase ? getCodeBase().toExternalForm() : getDocumentBase().toExternalForm();

        int idxSep  = base.lastIndexOf("/");
        if (idxSep != -1)
        {
          base = base.substring(0, idxSep + 1);
        }

        url = base + url;
      }
    }
  }

  /*****************************************************************************
   * Indicates whether a url is relative or absolute.
   *
   * @param   uri
   * @return True if the content reference is absolute.
   */
  private boolean isAbsolute(String uri)
  {
    boolean isAbsolute = false;

    // If the uri starts with / or \ marking an absolute path, or if it starts with a drive identifier with c: etc
    //
    if       (uri.startsWith("/") || uri.startsWith("\\"))                                   isAbsolute = true;
    else if  (uri.length() > 1 && uri.charAt(1) == ':' && Character.isLetter(uri.charAt(0))) isAbsolute = true;

    return isAbsolute;
  }

  /****************************************************************************
   * This method is just to display the sample applet player string
   * before we add the player renderer component, i.e.  before we Open any
   * content.  or after the component is removed which is after when we Close
   * the content (i.e when the player is not added to the frame)
   *
   * @param g      Graphics context
   */
  public void paint(Graphics g)
  {
    if (!playerAdded)
    {
      Rectangle r = getBounds();
      int w = r.width;
      int h = r.height;

      g.setColor(Color.black);
      g.fillRect(0, 0, w, h);

      Insets i = getInsets();
      w -= i.left + i.right;
      h -= i.top  + i.bottom;

      g.setFont(logoFont);
      FontMetrics fm = g.getFontMetrics();
      int sWidth = fm.stringWidth(kPlayerName);
      int iLeft  = i.left + (w - sWidth)/2;
      int y      = i.top  + h/2;
      g.setColor(Color.blue);
      g.drawString(kPlayerName, iLeft, y);
    }
  }
  //
  private Font logoFont = new Font("Helvetica", Font.BOLD, 12);

  // -------- PLAYER LISTENER ADAPTER TO LISTEN TO PLAYER EVENTS ------------------
  //
  // This is just to show how its done - its not all that useful to show the status
  // like this but we can override any of the methods in PlayerAdapter and use the
  // information in whatever manner suits the applet.
  //
  class M4AppletSampleAdapter extends PlayerAdapter
  {
    /******************************************************************************
     * Indicates change of state
     *
     * @param state  State as per PlayerControl
     *
     * @see player.PlayerControl
     */
    public void changedState(int state)
    {
      String stateMsg = "?";

      switch (state)
      {
        case PlayerControl.CLOSED:  stateMsg = "Closed";         break;
        case PlayerControl.PAUSED:  stateMsg = "Paused";         break;
        case PlayerControl.PLAYING: stateMsg = "Playing " + url; break;
        case PlayerControl.STOPPED: stateMsg = "Stopped";        break;
        default:                    stateMsg = "!";              break;
      }

      showStatus("M4AppletSample: " + stateMsg);
    }
  }

}
